/// <reference path="lodash.d.ts" />
/// <reference path="three.d.ts" />
/// <reference path="Utils.ts" />
/// <reference path="OrbitalMechanics.ts" />
/// <reference path="Rings.ts" />
/// <reference path="SOITree.ts" />
/// <reference path="SolarSystem.ts" />
/// <reference path="Labels.ts" />
// Body.ts
var Body = (function () {
    function Body(name, data, kind) {
        if (kind === void 0) { kind = "planet"; }
        this.name = name;
        this.data = data;
        this.kind = kind;
        this.material = null;
        this.radius = data.radius;
        this.orbitData = data.orbit;
        this.rotationData = data.rotation;
        this.color = HexCol(data.color);
        this.material = data.material || null;
        this.orbit = this.orbitData ? new Orbit(this.orbitData) : null;
        this.rotation = this.rotationData ? new Rotation(this.rotationData) : null;
    }
    Body.prototype.toString = function () {
        return "[Body name=" + this.name + ", radius=" + (this.radius * 1e-3).toFixed(2) + "km]";
    };
    return Body;
}());
/**
 * Responsible for rendering one SOI level.
 */
var LevelRenderer = (function () {
    function LevelRenderer(node, treeRenderer, scene) {
        if (scene === void 0) { scene = treeRenderer.scene; }
        this.node = node;
        this.treeRenderer = treeRenderer;
        this.scene = scene;
        console.log("  --- new LevelRenderer");
        this._setUp = false;
    }
    LevelRenderer.prototype.setUp = function () {
        var _this = this;
        if (this._setUp)
            console.error("setUp called twice!", this);
        console.log("    {{{ LevelRenderer::setUp", this.node.data);
        this.getNodes().forEach(function (node) {
            if (node.storage._nodeRenderer) {
                console.info("This node already have a central mesh.", _this, node);
                return;
            }
            var nodeRenderer = _this.makeNodeRenderer(node);
            nodeRenderer.setUp();
            node.storage._nodeRenderer = nodeRenderer;
        });
        this.node.storage._levelRenderer = this;
        this._setUp = true;
        console.log("    }}} LevelRenderer::setUp");
    };
    LevelRenderer.prototype.tearDown = function () {
        if (!this._setUp)
            console.error("tearDown called twice!", this);
        console.log("    {{{ LevelRenderer::tearDown", this.node.data);
        this.node.children.forEach(function (child) {
            child.storage._nodeRenderer.tearDown();
            child.storage._nodeRenderer = null;
        });
        this.node.storage._levelRenderer = null;
        this._setUp = false;
        console.log("    }}} LevelRenderer::tearDown");
    };
    LevelRenderer.prototype.makeNodeRenderer = function (node) {
        // choose a constructor based on node.data
        if (node.data instanceof Body) {
            var body = node.data;
            return new SphericalNodeRenderer(node, body, this);
        }
        else {
            console.error("node.data is not Body", node);
        }
    };
    /*
     * Called by the cursor to signal position change.
     * curPos is the position of the cursor relative to the SOINode.
     */
    LevelRenderer.prototype.onUpdatePosition = function (curPos) {
        this.node.storage._nodeRenderer.update(V3(), curPos);
        this.node.children.map(function (child) {
            child.storage._nodeRenderer.update(child.relPos, curPos);
        });
    };
    LevelRenderer.prototype.getNodes = function () {
        var xs = _.clone(this.node.children);
        xs.push(this.node);
        return xs;
    };
    return LevelRenderer;
}());
var SphericalNodeRenderer = (function () {
    function SphericalNodeRenderer(node, body, levelRenderer, scene) {
        if (scene === void 0) { scene = levelRenderer.scene; }
        this.node = node;
        this.body = body;
        this.levelRenderer = levelRenderer;
        this.scene = scene;
        this._setUp = false;
        this.angularDiameters = [0.5 * deg, 2 * deg, 8 * deg, Infinity];
        this.details = [6, 16, 64, 128];
        this.line = null;
        this.meshes = [];
        this.currentLOD = -1;
        this.label = LabelManager.instance.createLabel();
        this.label.setText(this.body.name);
        this._sceneTargets = this.scene._targets;
    }
    SphericalNodeRenderer.prototype._ensureOrbitLine = function () {
        if (this.line)
            return;
        var body = this.body;
        if (body.kind == "planet" || body.kind == "moon") {
            var orbit = body.orbit.makePlot();
            this.line = orbit.line;
            this.line.name = "Orbit Line: " + this.body.name;
            this.line.position.set(0, 0, 0);
            this.line.scale.set(1, 1, 1);
            this.line.quaternion.setFromAxisAngle(V3(1, 0, 0), 0);
            this.updateOrbit = orbit.updateGeom;
            this.scene._orbitLines.add(this.line);
            this.updateOrbit();
        }
    };
    SphericalNodeRenderer.prototype._updatePosition = function () {
        var orbit = this.body.orbit;
        if (orbit) {
            var t = this.node.soiTree.time;
            this.node.relPos.copy(orbit.getPos(t));
        }
        //if (this.updateOrbit) {
        //  this.updateOrbit();
        //}
    };
    SphericalNodeRenderer.prototype._updateRotation = function (mesh) {
        var rotation = this.body.rotation;
        if (rotation) {
            var t = this.node.soiTree.time;
            mesh.quaternion.copy(rotation.getRotation(t));
        }
    };
    /**
     * Update the sphere's position based on this object's displacement to the
     * cursor (displacement is actually position in WebGL world space).
     */
    SphericalNodeRenderer.prototype.update = function (relPos, cursorPos) {
        if (this.currentLOD == -1)
            console.error(this.body.name + ": currentLOD == -1", this);
        if (!this._setUp)
            console.error(this.body.name + ": _setUp is false!", this);
        var displacement = V3().subVectors(relPos, cursorPos);
        var maxDistance = 100 * au, minTheta = this.body.kind == "star" ? 0.006 : 0.003;
        var minThetaRaycastGeom = 0.01;
        var radius = this.body.radius;
        var distance = displacement.length();
        // angular radius on screen in radians
        var theta = radius / distance;
        // Update LOD
        var newLevel = 0;
        while (newLevel < this.meshes.length && this.meshes[newLevel].angularDiameter < 2 * theta)
            newLevel++;
        if (newLevel >= this.meshes.length)
            newLevel = this.meshes.length - 1;
        if (newLevel != this.currentLOD) {
            if (this.meshes[newLevel].mesh == null) {
                console.log(this.body.name + ": Requesting LOD " + newLevel, this);
                this.ensureLOD(newLevel);
            }
            else {
                console.log(this.body.name + ": Switch LOD from "
                    + this.currentLOD + " to " + newLevel +
                    ", angularDiameter:" + (2 * theta * 180 / Math.PI).toFixed(2) + " deg" +
                    ", radius:" + (radius * 1e-3).toFixed(2) + " km" +
                    ", distance:" + (distance * 1e-3).toFixed(2) + " km", this);
                this.deselect(this.currentLOD);
                this.select(newLevel);
            }
        }
        var mesh = this.meshes[this.currentLOD].mesh;
        if (mesh == null) {
            console.error(this.body.name + ": mesh at " + this.currentLOD + " is null!", this);
            return;
        }
        this._updatePosition();
        this._updateRotation(mesh);
        // Calculate new mesh position
        var actualDistance = distance;
        mesh.position.copy(displacement);
        // Set position of orbit renderer
        // displacement = relPos - cursorPos
        // linePos = displacement - relPos = (relPos - cursorPos) - relPos
        //                                 = -cursorPos
        if (this.line)
            this.line.position.copy(cursorPos).negate();
        // radius to preserve angular diameter while changing distance
        var scaledRadius = theta * actualDistance;
        // set mesh radius to fit angular diameter
        var size = Math.max(minTheta * actualDistance, scaledRadius);
        mesh.scale.set(size, size, size);
        // set raycast geometry scale
        var raycastGeom = mesh._raycastGeometry;
        var objTheta = size / actualDistance;
        var raycastGeomScale = objTheta > minThetaRaycastGeom ? 1 : minThetaRaycastGeom / objTheta;
        raycastGeom.scale.set(raycastGeomScale, raycastGeomScale, raycastGeomScale);
        var light = mesh._light;
        if (light) {
            light.lookAt(V3());
        }
        mesh.updateMatrix();
        mesh.updateMatrixWorld(false);
        this.label.reposition(mesh);
        var otherPos = V3().copy(this.levelRenderer.treeRenderer.cursor.localPos);
        var relPos = this.levelRenderer.treeRenderer.cursor.cursor.relativePositionTo(this.node);
        var dist = otherPos.add(relPos).length();
        this.label.setText(this.body.name); // + " (" + formatDistance(dist, 2) + ")");
    };
    SphericalNodeRenderer.prototype.setUp = function () {
        this._setUp = true;
        this.meshes = _.zipWith(this.angularDiameters, this.details, function (angularDiameter, detail) {
            return { angularDiameter: angularDiameter, detail: detail, mesh: null };
        });
        this._ensureOrbitLine();
        this.ensureLOD(0);
        this.select(0);
    };
    SphericalNodeRenderer.prototype.tearDown = function () {
        this.deselect(this.currentLOD);
        for (var i = 0; i < this.meshes.length; i++)
            this.removeLOD(i);
        this._setUp = false;
        this.label.remove();
        if (this.line) {
            this.scene._orbitLines.remove(this.line);
            this.updateOrbit = null;
            this.line = null;
        }
    };
    SphericalNodeRenderer.prototype.ensureLOD = function (index) {
        if (this.meshes[index].mesh != null)
            return;
        var lod = this.meshes[index];
        var detail = lod.detail;
        var geometry = new THREE.SphereGeometry(1, detail, detail);
        var material;
        if (detail < 10 || this.body.kind == "star") {
            var farColor = Col().copy(this.body.color).offsetHSL(0, -0.1, 0.1);
            material = new THREE.MeshBasicMaterial({
                color: farColor.getHex()
            });
        }
        else {
            if (this.body.material != null)
                material = this.body.material;
            else
                material = new THREE.MeshPhongMaterial({
                    color: this.body.color.getHex(),
                    specular: 0x222222,
                    shininess: 20
                });
        }
        var mesh = new THREE.Mesh(geometry, material);
        if (this.body.data.rings && detail > 10) {
            var ringsMesh = MakeRingMesh(this.body.data.rings);
            mesh.add(ringsMesh);
            ringsMesh.name = this.body.name + ": Rings";
            ringsMesh.position.set(0, 0, 0);
            var scale = 1 / this.body.data.radius;
            ringsMesh.scale.set(scale, scale, scale);
        }
        mesh.name = "SphericalNodeRenderer: " + this.body.toString();
        mesh._renderer = this;
        lod.mesh = mesh;
        if (this.body.kind == "star") {
            // For far distances, the sun is a directional light, but in this
            // game, there are no situations where the arrangement bodies where
            // the directional light approximation breaks down and point light is
            // needed. (things behind the sun have low LOD so they are not lighted)
            var light = new THREE.DirectionalLight(0xffffff, 0.8);
            mesh.add(light);
            mesh._light = light;
        }
        var raycastGeom = new THREE.Mesh(new THREE.SphereGeometry(2, 6, 6), new THREE.MeshBasicMaterial({
            color: 0xffffff, opacity: 0, transparent: true, depthWrite: false
        }));
        raycastGeom.name = this.body.name + ": raycastGeom";
        mesh._raycastGeometry = raycastGeom;
        raycastGeom._sphericalNodeRenderer = this;
        mesh.add(raycastGeom);
        this._sceneTargets.push(raycastGeom);
        console.log(this.body.name + ": ensureLOD: LOD instantiated: " +
            index + " (detail=" + detail + ")", lod, this);
    };
    SphericalNodeRenderer.prototype.removeLOD = function (index) {
        var lod = this.meshes[index];
        if (lod && lod.mesh) {
            var raycastGeom = lod.mesh._raycastGeometry;
            _.remove(this._sceneTargets, function (t) { return t === raycastGeom; });
        }
        lod.mesh = null;
        this.meshes[index] = null;
    };
    SphericalNodeRenderer.prototype.select = function (index) {
        if (index == -1)
            return;
        this.scene.add(this.meshes[index].mesh);
        this.currentLOD = index;
    };
    SphericalNodeRenderer.prototype.deselect = function (index) {
        if (index == -1)
            return;
        var mesh = this.meshes[index].mesh;
        mesh._renderer = null;
        this.scene.remove(mesh);
        this.currentLOD = -1;
    };
    return SphericalNodeRenderer;
}());
var SOITreeRenderer = (function () {
    function SOITreeRenderer(tree, scene, cursor) {
        var _this = this;
        if (cursor === void 0) { cursor = new SOICursor(tree, tree.root); }
        this.tree = tree;
        this.scene = scene;
        this.cursor = cursor;
        this.cursor.onMoveToChild(function (ch) {
            _this.pushRenderer(ch);
        });
        this.cursor.onMoveToParent(function (p) {
            _this.popRenderer();
        });
        this.renderers = [];
        this.pushRenderer(this.tree.root);
    }
    SOITreeRenderer.prototype.pushRenderer = function (node) {
        console.log("{{{ pushRenderer", node.data);
        var r = this.getLevel(node);
        r.setUp();
        this.renderers.push(r);
        console.log("}}} pushRenderer", this.renderers);
    };
    SOITreeRenderer.prototype.popRenderer = function () {
        console.log("{{{ popRenderer");
        var r = this.getTopRenderer();
        r.tearDown();
        this.renderers.pop();
        console.log("}}} popRenderer", this.renderers);
    };
    SOITreeRenderer.prototype.getTopRenderer = function () {
        return this.renderers[this.renderers.length - 1];
    };
    SOITreeRenderer.prototype.getLevel = function (node) {
        if (node.storage._levelRenderer)
            return node.storage._levelRenderer;
        else
            return new LevelRenderer(node, this);
    };
    SOITreeRenderer.prototype.moveCursor = function (delta) {
        this.cursor.move(delta);
        // relative position of current node, which computed by doing is partial sum
        // from children to root
        var curRelPos = V3().copy(this.cursor.localPos);
        _(this.renderers).forEachRight(function (renderer_) {
            var renderer = renderer_;
            renderer.onUpdatePosition(V3().copy(curRelPos));
            curRelPos.add(renderer.node.relPos);
        });
    };
    SOITreeRenderer.prototype.getRelativePositionTo = function (node) {
        return this.cursor.cursor.relativePositionTo(node).add(this.cursor.localPos);
    };
    return SOITreeRenderer;
}());
function getSolarSystem() {
    var solarSystemData = getSolarSystemData();
    var solarSystemPairs = _.toPairs(solarSystemData);
    var solarSystemRadius = 1.2 * _(solarSystemPairs).map(function (kv) { return kv[1].orbit.semiMajorAxis; }).max();
    var sun = new Body("Sun", { radius: 695500 * km, color: 0xffff00 }, "star");
    var solarSystem = new SOITree(solarSystemRadius, sun);
    var solarSystemItems = {};
    var planets = _.filter(solarSystemPairs, function (kv) { return kv[1].parent == null; });
    var moons = _.filter(solarSystemPairs, function (kv) { return kv[1].parent != null; });
    [planets, moons].forEach(function (arr) {
        arr.forEach(function (kv) {
            var name = kv[0], data = kv[1];
            var isMoon = data.parent != null;
            var color = data.color;
            var body = new Body(name, data, isMoon ? "moon" : "planet");
            var soiRadius = isMoon ? data.radius * 25 : data.radius * 3500;
            var theta = data.orbit.phase, R = data.orbit.semiMajorAxis;
            var pos = V3(Math.cos(theta) * R, 0, Math.sin(theta) * R);
            var parent = (isMoon ? solarSystemItems[data.parent] : solarSystem.root);
            var node = parent.addChild(pos, soiRadius, body);
            solarSystemItems[name] = node;
        });
    });
    function testRelativePositionTo() {
        var moon = solarSystemItems.Moon;
        var earth = solarSystemItems.Earth;
        var mars = solarSystemItems.Mars;
        var jupiter = solarSystemItems.Jupiter;
        var saturn = solarSystemItems.Saturn;
        console.log(moon.getRelativePositionList());
        // should be zero vector
        console.log("moon/moon", moon.relativePositionTo(moon));
        // should be negation of earth other
        console.log("moon/earth", moon.relativePositionTo(earth), "earth/moon", earth.relativePositionTo(moon));
        console.log("moon/sun", moon.relativePositionTo(solarSystem.root), "sun/moon", solarSystem.root.relativePositionTo(moon));
        // chain addition property of relative positions
        var moonEarth = moon.relativePositionTo(earth);
        var earthMars = earth.relativePositionTo(mars);
        var marsJupiter = mars.relativePositionTo(jupiter);
        var jupiterSaturn = jupiter.relativePositionTo(saturn);
        var sum = V3().add(moonEarth).add(earthMars).add(marsJupiter).add(jupiterSaturn);
        console.log("moon/saturn", moon.relativePositionTo(saturn), "= moon/earth", moonEarth, "+ earth/mars", earthMars, "+ mars/jupiter", marsJupiter, "+ jupiter/saturn", jupiterSaturn, "= moon/saturn", sum);
    }
    //testRelativePositionTo();
    return solarSystem;
}
//# sourceMappingURL=Body.js.map