/// <reference path="lodash.d.ts" />
/// <reference path="three.d.ts" />
/// <reference path="Utils.ts" />
// SOITree.ts
/**
 * Sphere of Influence Tree:
 *  Divides the world into a hierarchy of nodes, each node has a sphere of
 *  influence radius. When a point is inside a tree, the closest SOI tree node
 *  is determined.
 */
var SOITree = (function () {
    function SOITree(rootRadius, rootData) {
        if (rootData === void 0) { rootData = null; }
        this.time = 0.0;
        this.root = new SOINode(this, null, V3(), rootRadius, [], rootData);
    }
    return SOITree;
}());
var SOINode = (function () {
    function SOINode(soiTree, parent, relPos, radius, children, 
        /** Data associated to this SOINode. */
        data, 
        /** Storage for other classes to modify. */
        storage) {
        if (data === void 0) { data = {}; }
        if (storage === void 0) { storage = {}; }
        this.soiTree = soiTree;
        this.parent = parent;
        this.relPos = relPos;
        this.radius = radius;
        this.children = children;
        this.data = data;
        this.storage = storage;
    }
    SOINode.prototype.getParentPosition = function (pos, into) {
        if (into === void 0) { into = null; }
        if (into == null)
            into = V3();
        if (this.parent == null)
            return into.copy(pos);
        else
            return into.addVectors(this.relPos, pos);
    };
    SOINode.prototype.getGlobalPosition = function (into) {
        if (into === void 0) { into = null; }
        if (into == null)
            into = V3();
        if (this.parent == null) {
            return into;
        }
        else {
            into.add(this.relPos);
            return this.parent.getGlobalPosition(into);
        }
    };
    SOINode.prototype.addChild = function (relPos, radius, data) {
        if (data === void 0) { data = null; }
        var newChild = new SOINode(this.soiTree, this, relPos, radius, [], data);
        this.children.push(newChild);
        return newChild;
    };
    SOINode.prototype.contains = function (otherRelPos) {
        return otherRelPos.lengthSq() <= this.radius * this.radius;
    };
    SOINode.prototype.containsByParentPos = function (otherRelPos) {
        return this.relPos.distanceToSquared(otherRelPos) <= this.radius * this.radius;
    };
    /**
     * Choose a new child by finding a closer object.
     * parentPos is the position relative to parent. For example, if parent is
     * Earth, then parentPos is relative to Earth and the Moon is chosen if the
     * ship is closer to Moon than Earth.
     */
    SOINode.prototype.findChildByParentPos = function (parentPos) {
        var centralD2 = parentPos.lengthSq();
        var res = _(this.children)
            .filter(function (child) { return child.containsByParentPos(parentPos)
            && child.relPos.distanceToSquared(parentPos) <= centralD2; })
            .minBy(function (child) { return child.relPos.distanceToSquared(parentPos); });
        return res || null;
    };
    SOINode.prototype.getRelativePositionList = function () {
        var pos = V3(), node = this;
        var nodes = [];
        while (node) {
            nodes.push({ node: node, pos: V3().copy(pos), name: node.data ? node.data.name : "" });
            pos.add(node.relPos);
            node = node.parent;
        }
        return nodes;
    };
    SOINode.prototype.relativePositionTo = function (node) {
        // the relative pos between this and node is
        // absPos(this) - absPos(node), or relPos(this, c) - relPos(this, c)
        // for nearest ancestor c
        var left = this.getRelativePositionList();
        var other = node.getRelativePositionList();
        for (var i = 0; i < left.length; i++) {
            for (var j = 0; j < other.length; j++) {
                if (left[i] === other[j]) {
                    return V3().subVectors(left[i].pos, other[j].pos);
                }
            }
        }
        return V3().subVectors(left[left.length - 1].pos, other[other.length - 1].pos);
    };
    SOINode.prototype.sqDistanceTo = function (node) {
        return this.relativePositionTo(node).lengthSq();
    };
    SOINode.prototype.distanceTo = function (node) {
        return this.relativePositionTo(node).length();
    };
    return SOINode;
}());
var SOICursor = (function () {
    function SOICursor(soiTree, cursor, localPos) {
        if (localPos === void 0) { localPos = V3(); }
        this.soiTree = soiTree;
        this.cursor = cursor;
        this.localPos = localPos;
        this._onMoveToParent = [];
        this._onMoveToChild = [];
        this.onMoveToParent(function (p) { return console.log("onMoveToParent", p); });
        this.onMoveToChild(function (p) { return console.log("onMoveToChild", p); });
    }
    SOICursor.prototype.move = function (deltaPos) {
        /*
         * Compute new relative position.
         * Is object still in cursor SOI?
         * No:
         *   Move to parent SOI
         * Yes:
         *   Is object going to enter a child SOI?
         *     Yes:
         *       Enter to child SOI
         *
         */
        this.localPos.add(deltaPos);
        if (!this.cursor.contains(this.localPos)) {
            console.log("Move to parent.", this, this.cursor.parent);
            this.moveToParent();
            return -1;
        }
        else {
            var result = this.cursor.findChildByParentPos(this.localPos);
            if (result !== null) {
                this.moveToChild(result);
                return 1;
            }
        }
        return 0;
    };
    SOICursor.prototype.getGlobalPosition = function (into) {
        if (into === void 0) { into = null; }
        return this.cursor.getGlobalPosition(into).add(this.localPos);
    };
    SOICursor.prototype.moveToParent = function () {
        var _this = this;
        if (this.cursor.parent === null)
            return false;
        this.localPos.add(this.cursor.relPos);
        this.cursor = this.cursor.parent;
        this._onMoveToParent.forEach(function (f) { return f(_this.cursor); });
        return true;
    };
    SOICursor.prototype.moveToChild = function (node) {
        var _this = this;
        if (node == null || node.parent != this.cursor)
            return false;
        this.localPos.sub(node.relPos);
        this.cursor = node;
        this._onMoveToChild.forEach(function (f) { return f(_this.cursor); });
        return true;
    };
    SOICursor.prototype.onMoveToParent = function (f) {
        this._onMoveToParent.push(f);
    };
    SOICursor.prototype.onMoveToChild = function (f) {
        this._onMoveToChild.push(f);
    };
    return SOICursor;
}());
//# sourceMappingURL=SOITree.js.map