/// <reference path="lodash.d.ts" />
/// <reference path="three.d.ts" />
/// <reference path="Utils.ts" />

// SOITree.ts

/**
 * Sphere of Influence Tree:
 *  Divides the world into a hierarchy of nodes, each node has a sphere of
 *  influence radius. When a point is inside a tree, the closest SOI tree node
 *  is determined.
 */

class SOITree {
  public root : SOINode;
  public time = 0.0;

  constructor(rootRadius : number, rootData = null as any) {
    this.root = new SOINode(this, null, V3(), rootRadius, [], rootData);
  }
}

class SOINode {
  constructor(
    public soiTree : SOITree,
    public parent : SOINode,
    public relPos : Vector3,
    public radius : number,
    public children : SOINode[],
    /** Data associated to this SOINode. */
    public data = {} as any,
    /** Storage for other classes to modify. */
    public storage = {} as any
  ) { }

  public getParentPosition(pos : Vector3, into = null as Vector3) {
    if (into == null) into = V3();

    if (this.parent == null)
      return into.copy(pos);
    else
      return into.addVectors(this.relPos, pos);
  }

  public getGlobalPosition(into = null as Vector3) : Vector3 {
    if (into == null) into = V3();

    if (this.parent == null) {
      return into;
    } else {
      into.add(this.relPos);
      return this.parent.getGlobalPosition(into);
    }
  }

  public addChild(relPos : Vector3, radius : number, data = null as any) {
    var newChild = new SOINode(this.soiTree, this, relPos, radius, [], data);
    this.children.push(newChild);
    return newChild;
  }

  public contains(otherRelPos : Vector3) : boolean {
    return otherRelPos.lengthSq() <= this.radius * this.radius;
  }

  public containsByParentPos(otherRelPos : Vector3) : boolean {
    return this.relPos.distanceToSquared(otherRelPos) <= this.radius * this.radius;
  }

  /**
   * Choose a new child by finding a closer object.
   * parentPos is the position relative to parent. For example, if parent is
   * Earth, then parentPos is relative to Earth and the Moon is chosen if the
   * ship is closer to Moon than Earth.
   */
  public findChildByParentPos(parentPos : Vector3) {
    var centralD2 = parentPos.lengthSq();
    var res = _(this.children)
      .filter((child) => child.containsByParentPos(parentPos)
                         && child.relPos.distanceToSquared(parentPos) <= centralD2)
      .minBy((child) => child.relPos.distanceToSquared(parentPos));

    return res || null;
  }

  public getRelativePositionList() : [{pos : Vector3, node : SOINode, name : string}] {
    var pos = V3(), node : SOINode = this;
    var nodes = [];
    while (node) {
      nodes.push({node: node, pos: V3().copy(pos), name: node.data ? node.data.name : ""});
      pos.add(node.relPos);
      node = node.parent;
    }
    return nodes as any;
  }

  public relativePositionTo(node : SOINode) : Vector3 {
    // the relative pos between this and node is
    // absPos(this) - absPos(node), or relPos(this, c) - relPos(this, c)
    // for nearest ancestor c
    var left = this.getRelativePositionList();
    var other = node.getRelativePositionList();

    for (var i = 0; i < left.length; i++) {
      for (var j = 0; j < other.length; j++) {
        if (left[i] === other[j]) {
          return V3().subVectors(left[i].pos, other[j].pos);
        }
      }
    }

    return V3().subVectors(left[left.length - 1].pos, other[other.length - 1].pos);
  }

  public sqDistanceTo(node : SOINode) : number {
    return this.relativePositionTo(node).lengthSq();
  }

  public distanceTo(node : SOINode) : number {
    return this.relativePositionTo(node).length();
  }
}

class SOICursor {
  private _onMoveToParent : ((p : SOINode) => void)[];
  private _onMoveToChild : ((p : SOINode) => void)[];

  constructor(
    public soiTree : SOITree,
    public cursor : SOINode,
    public localPos = V3()
  ) {
    this._onMoveToParent = [];
    this._onMoveToChild = [];

    this.onMoveToParent((p) => console.log("onMoveToParent", p));
    this.onMoveToChild((p) => console.log("onMoveToChild", p));
  }

  public move(deltaPos : Vector3) {
    /*
     * Compute new relative position.
     * Is object still in cursor SOI?
     * No:
     *   Move to parent SOI
     * Yes:
     *   Is object going to enter a child SOI?
     *     Yes:
     *       Enter to child SOI
     *
     */

    this.localPos.add(deltaPos);
    if (!this.cursor.contains(this.localPos)) {
      console.log("Move to parent.", this, this.cursor.parent);
      this.moveToParent();
      return -1;
    } else {
      var result = this.cursor.findChildByParentPos(this.localPos);
      if (result !== null) {
        this.moveToChild(result);
        return 1;
      }
    }
    return 0;
  }

  public getGlobalPosition(into = null as Vector3) : Vector3 {
    return this.cursor.getGlobalPosition(into).add(this.localPos);
  }

  private moveToParent() {
    if (this.cursor.parent === null)
      return false;

    this.localPos.add(this.cursor.relPos);
    this.cursor = this.cursor.parent;

    this._onMoveToParent.forEach(f => f(this.cursor));
    return true;
  }

  private moveToChild(node : SOINode) {
    if (node == null || node.parent != this.cursor)
      return false;

    this.localPos.sub(node.relPos);
    this.cursor = node;

    this._onMoveToChild.forEach(f => f(this.cursor));
    return true;
  }

  public onMoveToParent(f) {
    this._onMoveToParent.push(f);
  }

  public onMoveToChild(f) {
    this._onMoveToChild.push(f);
  }
}

